!============================================================
! sdkLCS.clw -- List Column Sorter class implementation
!
!  2010 Simon Kemp, simon@sdk.demon.co.uk
!
!  This software is provided 'as-is', without any express or implied
!  warranty.  In no event will the author be held liable for any damages
!  arising from the use of this software.
!
!  Permission is granted to anyone to use this software for any purpose,
!  including commercial applications, and to alter it and redistribute it
!  freely, subject to the following restrictions:
!
!  1. The origin of this software must not be misrepresented; you must not
!     claim that you wrote the original software. If you use this software
!     in a product, an acknowledgment in the product documentation would be
!     appreciated but is not required.
!
!  2. Altered source versions must be plainly marked as such, and must not be
!     misrepresented as being the original software.
!
!  3. This notice may not be removed or altered from any source distribution.
!
!============================================================
    member()
    include('sdklcs.inc'),once
    include('keycodes.clw'),once

    map
      SDKLCSCompareFunc(*group, *group),signed
      module('clarionrtl')
        SDKLCSSortFreeRTL(*queue, *SDKLCSCompareFunc),NAME('Cla$FREEQueueFKey')
      end
    end

sdkLCS.Init  procedure(long thelistfeq, *queue listqueue, long tweaklineheight=0)

    code

    free(self.columns)
    free(self.keys)
    self.alertkeystd = MouseLeft
    self.alertkeymulti = CtrlMouseLeft
    self.alertkeyclear = ShiftMouseLeft
    self.headereffects = false
    self.multikey = true
    self.prependindicators = false
    self.showcursoronsort = true
    self.sortonchanges = true
    self.listfeq = thelistfeq
    self.listfrom &= listqueue
    if tweaklineheight <> 0
      self.listfeq{PROP:LineHeight} = self.listfeq{PROP:LineHeight} + tweaklineheight
    end
    self.listfeq{PROP:Alrt, 255} = self.alertkeystd
    self.listfeq{PROP:Alrt, 255} = self.alertkeymulti
    self.listfeq{PROP:Alrt, 255} = self.alertkeyclear
    self.stickysort &= null
    clear(self.columns)
    loop
      self.columns.column += 1
      if ~self.listfeq{PROPLIST:Exists, self.columns.column}
        break
      end
      self.columns.headertext = self.listfeq{PROPLIST:Header, self.columns.column}
      self.columns.fieldno = self.listfeq{PROPLIST:FieldNo, self.columns.column}
      self.columns.columnwho = who(self.listfrom, self.columns.fieldno)
      self.columns.isastring = isstring(what(self.listfrom, self.columns.fieldno))
      add(self.columns)
    end
    self.blockcolumnoptions = false
    if self.regionfeq = 0
      self.regionfeq = create(0,CREATE:region)
      self.regionfeq{prop:imm} = true
    end


sdkLCS.Kill  procedure

    code

    free(self.keys)
    self.ShowHeaders
    free(self.columns)
    self.listfrom &= null


sdkLCS.SetColumnOptions  procedure(<string nosorting>, <string casesensitive>, <string restrictsort>, <string overridecolumnfields>)

myPtr1  long
myAny   any

    code

    if self.blockcolumnoptions = true
      return
    end
    self.blockcolumnoptions = true
    loop myPtr1 = 1 to len(clip(overridecolumnfields)) - 5 by 6
      self.columns.column = overridecolumnfields[myPtr1 : myPtr1 + 2]
      get(self.columns, +self.columns.column)
      if ~errorcode()
        myAny &= null
        myAny &= what(self.listfrom,overridecolumnfields[myPtr1 + 3 : myPtr1 + 5])
        if ~myAny &= null
          self.columns.fieldno = overridecolumnfields[myPtr1 + 3 : myPtr1 + 5]
          self.columns.columnwho = who(self.listfrom, self.columns.fieldno)
          self.columns.isastring = isstring(what(self.listfrom, self.columns.fieldno))
          put(self.columns)
        end
      end
    end
    loop self.columns.column = 1 to len(clip(nosorting))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if nosorting[self.columns.column] = true
        self.columns.sortoptions += 2^0 ! no sorting
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(casesensitive))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if casesensitive[self.columns.column] = true
        self.columns.sortoptions += 2^1 ! case-sensitive
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(restrictsort))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if inrange(restrictsort[self.columns.column],1,3)
        execute restrictsort[self.columns.column]
          self.columns.sortoptions += 2^2 ! restrict +
          self.columns.sortoptions += 2^3 ! restrict -
          begin
            if ~band(self.columns.sortoptions, 2^0)
              self.columns.sortoptions += 2^0 ! no sorting
            end
          end
        end
        put(self.columns)
      end
    end


sdkLCS.SetInitialSort  procedure(<*string variablesort>, <string initialsort>, long locked=0)

    code

    if ~omitted(variablesort)
      self.stickysort &= variablesort
      self.initialsort = left(variablesort)
    else
      self.initialsort = left(initialsort)
    end
    self.locked = locked


sdkLCS.Start procedure

myPtr1          long
myPtr2          long

    code

    self.blockcolumnoptions = true
    if self.initialsort
      case self.initialsort[1]
      of '+' orof '-'
      else
        self.initialsort = '+' & self.initialsort
      end
      myPtr1 = 1
      loop while myPtr1 < len(clip(self.initialsort))
        loop myPtr2 = myPtr1 + 1 to len(clip(self.initialsort)) + 1
          clear(self.keys)
          if ~inrange(val(self.initialsort[myPtr2]),val('0'),val('9'))
            self.keys.level = records(self.keys) + 1
            self.keys.column = deformat(self.initialsort[myPtr1 : myPtr2 - 1])
            if self.keys.column < 0
              self.keys.direction = -1
            else
              self.keys.direction = +1
            end
            self.keys.column = abs(self.keys.column)
            self.columns.column = self.keys.column
            get(self.columns,+self.columns.column)
            if ~errorcode() and ~band(self.columns.sortoptions,2^0) and ~(band(self.columns.sortoptions,2^2) and self.keys.direction < 0) and ~(band(self.columns.sortoptions,2^3) and self.keys.direction > 0)
              self.keys.fieldno = self.columns.fieldno
              self.keys.sortoptions = self.columns.sortoptions
              self.keys.isastring = self.columns.isastring
              add(self.keys)
            end
            break
          end
        end
        myPtr1 = myPtr2
      end
      if ~self.multikey
        loop while records(self.keys) > 1
          get(self.keys,records(self.keys))
          delete(self.keys)
        end
      end
    end
    if ~self.stickysort &= null
      self.stickysort = self.showsort()
    end
    self.ShowHeaders
    self.Sort


sdkLCS.ShowHeaders  procedure(long reset=0)

myPtr   long

    code

    loop myPtr = 1 to records(self.columns)
      get(self.columns, myPtr)
      if reset
        self.listfeq{PROPLIST:Header, self.columns.column} = self.columns.headertext
      else
        self.keys.column = self.columns.column
        get(self.keys, +self.keys.column)
        if errorcode()
          self.listfeq{PROPLIST:Header, self.columns.column} = self.columns.headertext
        elsif self.prependindicators
          self.listfeq{PROPLIST:Header, self.columns.column} = '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']' & left(self.columns.headertext)
        else
          self.listfeq{PROPLIST:Header, self.columns.column} = clip(self.columns.headertext) & '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']'
        end
      end
    end


sdkLCS.ShowSort  procedure ! ,string

myPtr       long
myResult    string(1024)

    code

    sort(self.keys, +self.keys.level)
    loop myPtr = 1 to records(self.keys)
      get(self.keys, myPtr)
      if self.keys.direction < 0
        myResult = clip(myResult) & '-' & self.keys.column
      else
        myResult = clip(myResult) & '+' & self.keys.column
      end
    end
    return clip(myResult)


sdkLCS.Sort procedure

mySaveQ &string
myPtr   long

    code

    if self.listfrom &= null
      return
    end
    mySaveQ &= new string(size(self.listfrom))
    if records(self.listfrom) and records(self.keys)
      if self.showcursoronsort
        setcursor(CURSOR:Wait)
      end
      myPtr = choice(self.listfeq)
      if myPtr
        get(self.listfrom, myPtr)
        mySaveQ = self.listfrom
      end
      sort(self.keys, +self.keys.level)
      SDKLCSkeyqref &= self.keys
      SDKLCSSortFreeRTL(self.listfrom, SDKLCSCompareFunc)
      sort(self.listfrom, SDKLCSCompareFunc)
      if myPtr
        loop myPtr = 1 to records(self.listfrom)
          get(self.listfrom, myPtr)
          if mySaveQ = self.listfrom
            self.listfeq{PROP:Selected} = myPtr
            break
          end
        end
      end
      if self.showcursoronsort
        setcursor
      end
    end
    self.Storechanges = changes(self.listfrom)
    dispose(mySaveQ)


sdkLCS.TakeEvent  procedure ! ,long

    code

    if self.listfrom &= null
      case field()
      of self.listfeq
        case event()
        of EVENT:PreAlertKey
          return true
        end
      end
      return false
    end
    if self.sortonchanges
      if self.storechanges <> changes(self.listfrom)
        self.Sort
        return false
      end
    end
    case field()
    of self.listfeq
      case event()
      of EVENT:PreAlertKey
        case keycode()
        of self.alertkeystd orof self.alertkeymulti orof self.alertkeyclear
          if ~self.locked
            if self.listfeq{PROPLIST:MouseDownRow} = 0 and self.listfeq{PROPLIST:MouseDownZone} = LISTZONE:header
              self.lastfield = self.listfeq{PROPLIST:MouseDownField}
              self.lastkeycode = keycode()
              if keycode() = self.alertkeyclear
                if self.lastfield
                  if self.headereffects = false or self.listfeq{PROP:ListFEQ} > 0 ! couldn't get prop:edit working with drop lists and combos
                    self.actionkeycode(keycode())
                  else
                    self.regiondown
                  end
                end
              else
                self.columns.column = self.lastfield
                get(self.columns, +self.columns.column)
                if ~errorcode() and ~band(self.columns.sortoptions,2^0)
                  if self.headereffects = false or self.listfeq{PROP:ListFEQ} > 0 ! couldn't get prop:edit working with drop lists and combos
                    self.actionkeycode(keycode())
                  else
                    self.regiondown
                  end
                else
                  clear(self.lastfield)
                end
              end
            end
          end
          return true
        end
      end
    of self.regionfeq
      if self.regionfeq{PROP:Hide}= false
        case event()
        of EVENT:MouseOut
          self.listfeq{PROP:Edit,self.lastfield} = 0
          self.regionfeq{PROP:Hide} = true
        of EVENT:MouseUp
          if self.regionfeq{PROP:Hide} = false
            self.listfeq{PROP:Edit,self.lastfield} = 0
            self.regionfeq{PROP:Hide} = true
            self.actionkeycode(self.lastkeycode)
          end
        end
      end
    end
    return false


sdkLCS.ActionKeyCode  procedure(long thekeycode) ! ,private

mysavesort  string(1024)
myptr       long

    code

    case thekeycode
    of self.alertkeystd orof self.alertkeymulti orof self.alertkeyclear
      mysavesort = self.ShowSort()
      if self.lastkeycode = self.alertkeyclear
        free(self.keys)
      else
        sort(self.keys, +self.keys.level)
        self.columns.column = self.lastfield
        get(self.columns, +self.columns.column)
        if ~errorcode() and ~band(self.columns.sortoptions,2^0)
          self.keys.column = self.columns.column
          get(self.keys, +self.keys.column)
          if errorcode()
            if ~self.multikey or keycode() = self.alertkeystd
              free(self.keys)
            end
            clear(self.keys)
            self.keys.column = self.columns.column
            self.keys.direction = 1
            self.keys.fieldno = self.columns.fieldno
            self.keys.sortoptions = self.columns.sortoptions
            self.keys.isastring = self.columns.isastring
            add(self.keys)
          else
            self.keys.direction *= -1
            if ~self.multikey or keycode() = self.alertkeystd
              free(self.keys)
              add(self.keys)
            else
              put(self.keys)
            end
          end
          if (band(self.keys.sortoptions,2^2) and self.keys.direction < 0) or (band(self.keys.sortoptions,2^3) and self.keys.direction > 0)
            self.keys.direction *= -1
            put(self.keys)
          end
        end
      end
      loop myPtr = 1 to records(self.keys)
        get(self.keys, myPtr)
        self.keys.level = myPtr
        put(self.keys)
      end
      if ~self.stickysort &= null
        self.stickysort = self.showsort()
      end
      if mysavesort <> self.ShowSort()
        self.ShowHeaders
        self.Sort
      end
    end


sdklcs.FieldInGroup  procedure ! ,long,private

lRow            long,auto
lGroup          long,auto
lIndex          long,auto
lLastColumns    long,auto

    code

    if self.listfeq{PROPLIST:GroupNo + PROPLIST:Group,self.lastfield} + 0 = 0
      return 0
    else
      lGroup = self.listfeq{PROPLIST:GroupNo,self.lastfield} + 0
      lRow   = 2
      lIndex = 0
      loop lIndex = self.lastfield to 1 by -1
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          lLastColumns = lIndex - 1
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
           lRow+=1
        end
      end
      lIndex = self.lastfield + 1
      loop
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          lLastColumns=lIndex-1
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
          lRow += 1
        end
        lIndex += 1
      end
      if self.listfeq{PROPLIST:LastOnLine,lLastColumns}
        lRow -= 1
      end
      if ~self.listfeq{PROPLIST:Group + PROPLIST:Header,self.lastfield}
        lRow -= 1
      end
      return lRow
    end


sdkLCS.RowSelectedInGroup  procedure ! ,long,private

lRow        long,auto
lGroup      long,auto
lIndex      long,auto
lColumns    long,auto

    code

    if self.listfeq{PROPLIST:GroupNo + PROPLIST:Group,self.lastfield} + 0 = 0
      return 0
    else
      lGroup = self.listfeq{PROPLIST:GroupNo,self.lastfield} + 0
      lRow = 2
      lIndex = 0
      if self.listfeq{PROPLIST:LastOnLine,self.lastfield}
        lRow = 1
      end
      loop lIndex = self.lastfield to 1 by -1
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
          lRow += 1
        end
      end
      if ~self.listfeq{PROPLIST:Group + PROPLIST:Header,self.lastfield}
        lRow -= 1
      end
      return lRow
    end


sdklcs.RegionDown  procedure

lPixels         long
lRow            long
lDelta          long
lScrollWidth    long
lOldControl     long
lx1             long
lx2             long

    code

    lPixels = 0{prop:pixels}
    0{prop:pixels} = true
    loop lx2 = self.listfeq{PROP:HScrollPos} + 1 to self.lastfield - 1
      lx1 += self.listfeq{PROPLIST:width,lx2}
    end
    self.regionfeq{PROP:XPos} = lx1
    self.regionfeq{PROP:Width} = self.listfeq{PROPLIST:width,self.lastfield}

    lOldControl = self.listfeq{PROP:Edit,self.lastfield}
    self.listfeq{PROP:Edit,self.lastfield} = self.regionfeq
    if lOldControl = self.regionfeq
      lOldControl = 0
    end
    self.listfeq{PROP:Edit,self.lastfield} = lOldControl

    lRow = self.RowSelectedInGroup()
    if lRow
      self.regionfeq{PROP:Height} = (self.listfeq{PROP:HeaderHeight} / self.FieldInGroup())
      self.regionfeq{PROP:YPos} = (lRow * self.regionfeq{PROP:Height})-self.regionfeq{PROP:Height}-1
      self.regionfeq{PROP:Height} = self.regionfeq{PROP:Height} + 1
    else
      self.regionfeq{PROP:Height} = self.listfeq{PROP:HeaderHeight} - 1
      self.regionfeq{PROP:YPos} = 0
    end
    if self.listfeq{PROPLIST:Icon,self.lastfield}
      lDelta = self.listfeq{PROPLIST:Width,self.lastfield} - self.regionfeq{PROP:Width}
      if lDelta<0
        lDelta = 15
      end
      self.regionfeq{PROP:XPos} = self.regionfeq{PROP:XPos} - lDelta
      self.regionfeq{PROP:Width} = self.regionfeq{PROP:Width} + lDelta
    end
    if self.listfeq{PROP:VScroll} and records(self.listfrom) > self.listfeq{PROP:Items}
      lScrollWidth = 20  ! getsystemmetrics perhaps......
    else
      lScrollWidth = 3
    end
    if (self.regionfeq{PROP:XPos}+self.regionfeq{PROP:Width}>=self.listfeq{PROP:Width}-lScrollWidth)
      self.regionfeq{PROP:Width}=self.regionfeq{PROP:Width}-(self.regionfeq{PROP:XPos}+self.regionfeq{PROP:Width}-(self.listfeq{PROP:Width}-lScrollWidth+1))
    end
    self.regionfeq{PROP:BevelInner} = -1
    self.regionfeq{PROP:BevelOuter} = -1
    self.regionfeq{PROP:YPos} = self.regionfeq{PROP:YPos}+self.listfeq{PROP:YPos}+1
    self.regionfeq{PROP:XPos} = self.regionfeq{PROP:XPos}+self.listfeq{PROP:XPos}+1
    self.regionfeq{PROP:Hide} = false
    0{prop:pixels} = lPixels


sdkLCS.Construct  procedure ! ,private

    code

    self.columns &= new sdkLCScolq
    self.keys &= new sdkLCSkeyq


sdkLCS.Destruct  procedure ! ,private

    code

    dispose(self.columns)
    dispose(self.keys)


SDKLCSCompareFunc procedure(*group cfTarget, *group cfCompare) ! ,signed

myPtr       long
myanyt      any
myanyc      any

    code

    loop myPtr = 1 to records(SDKLCSkeyqref)
      get(SDKLCSkeyqref, myPtr)
      myanyt &= what(cfTarget, SDKLCSkeyqref.fieldno)
      myanyc &= what(cfCompare, SDKLCSkeyqref.fieldno)
      if SDKLCSkeyqref.isastring and ~band(SDKLCSkeyqref.sortoptions,2^1)
        ! case-less string sort required
        if upper(myanyt) < upper(myanyc)
          return -SDKLCSkeyqref.direction
        elsif upper(myanyt) > upper(myanyc)
          return SDKLCSkeyqref.direction
        end
      elsif myanyt < myanyc
        return -SDKLCSkeyqref.direction
      elsif myanyt > myanyc
        return SDKLCSkeyqref.direction
      end
    end
    return 0 ! all keys are equal

